// Random Dot Stereogram creation & manipulation
// Copyright (c) 2023 Terence Kelly <tpkelly@acm.org>

// See "Drill Bits" column in July/August 2023 issue of
// ACM _Queue_ magazine:  https://queue.acm.org/DrillBits

// TODO:  consider unsigned char instead of char  

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include "ck.h"

#define ES "errno = %d => '%s'\n"
#define EA errno, strerror(errno)

// Adds header & copies random bytes from stdin to stdout to create
// random PBM file of given dimensions.  Format spec:  man 5 pbm.
// Random bytes on stdin can come from, e.g., /dev/random.
static void randpbm(int w, int h) {
  char *p;  size_t rl;
  CK(w > 0 && h > 0, "bad w %d or h %d", w, h);
  rl = (size_t)(w / 8) + !!(w % 8);  // row length in bytes for w bits
  CK(NULL != (p = (char *)malloc(rl)), ES, EA);
  printf("P4\n%d %d\n", w, h);
  while (h--) {
    CK(fread (p, sizeof *p, rl, stdin ) == rl, ES, EA);
    CK(fwrite(p, sizeof *p, rl, stdout) == rl, ES, EA);
  }
}

// mmap given file, return pointer to start and file size
static const char * mm(const char *file, size_t *L) {
  int fd;  struct stat s;  char *r;
  CK(0 < (fd = open(file, O_RDONLY)), ES, EA);
  CK(0 == fstat(fd, &s), ES, EA);
  *L = (size_t)s.st_size;
  CK(MAP_FAILED != (r = (char *)mmap(NULL, *L, PROT_READ, MAP_PRIVATE, fd, 0)), ES, EA);
  CK(0 == close(fd), ES, EA);
  return (char *)r;
}

// Get dimensions of PBM bitmap.  Header must resemble "P4\n47 13\n".
// Return header length.  Constrain width and height so that product
// can't overflow a signed int.
static size_t getdim(const char *f, const char *p, int *w, int *h) {
  int W, H, n = 0;
  CK(2 == sscanf(p, "P4 %d %d\n%n", &W, &H, &n), "f='%s' parse error", f   );
  CK(0 < n &&    16 > n,                         "f='%s' header size", f   );
  CK(0 < W && 40000 > W,                         "f='%s' width %d",    f, W);
  CK(0 < H && 40000 > H,                         "f='%s' height %d",   f, H);
  *w = W;
  *h = H;
  return (size_t)n;
}

// Return size in bytes of PBM raster.  Rows may contain trailing
// "don't care" bits; see man 5 pbm.
static size_t rs(const int w, const int h) {
  return (size_t)(((w / 8) + (!!(w % 8))) * h);
}

static int getbit(const char *a, const int w, const int h, const int r, const int c) {
  assert(r < h && c < w);
  a += (size_t)r * (rs(w, h) / (size_t)h);  // move to correct row
  a += c / 8;                               // move to correct column within row
  return (*a >> (7 - (c % 8))) & 0x1;       // extract packed bit from byte
}

// set bit in given row buffer
static void setbit(char *a, const int w, const int c, const int b) {
  char m;
  assert(c < w && (0 == b || 1 == b));
  a += c / 8;                               // move to correct column within row
  m = 0x1 << (7 - (c % 8));                 // find bit within byte
  if (b) *a |=  m;  // set
  else   *a &= ~m;  // clear
}

// Read random byte from stdin and return random bit.
// Input can come from, e.g., /dev/random.
static int randbit(void) {
  unsigned char b;
  CK(1 == fread(&b, 1, 1, stdin), ES, EA);
  return 127 < b;  // profligate with entropy:  collapses byte to bit
}

// Print to stdout random-dot stereogram based on given bitmaps P and K.
// P and K are names of PBM files of identical dimensions.  P contains
// the image that the RDS will encode; it can be created with the
// pbmtext command-line utility or, for more readable RDSes, with my custom
// font using a script like "msg.csh".  K is a ranom dot pattern; it can be
// created using my randpbm() function above.
static void makerds(const char *p, const char *k) {
  size_t pL, kL, phs, khs, rbs;  int pw, ph, kw, kh, spill = 0;
  const int D = 2;  // displacement:  pixels shifted right by D positions
  char *R;  // output row buffer
  const char *P = mm(p, &pL),
             *K = mm(k, &kL);
  phs = getdim(p, P, &pw, &ph);
  khs = getdim(k, K, &kw, &kh);
  CK(pL == kL,               "mismatch file sizes %lu %lu", pL, kL);  // "%z"?
  CK(pw == kw,               "mismatch widths %d %d", pw, kw);
  CK(ph == kh,               "mismatch heights %d %d", ph, kh);
  CK(pL - phs == rs(pw, ph), "mismatch dim/len %d %d => %lu", pw, ph, rs(pw, ph));
  printf("P4\n%d %d\n", pw, ph);  // print output header
  // allocate output row buffer with enough bytes for a row of bits,
  // allowing for "don't care" bits
  rbs = rs(pw, ph) / (size_t)ph;
  CK(NULL != (R = (char *)malloc(rbs)), ES, EA);
  P += phs;
  K += khs;
  // warn if shifted P spills over bitmap edge
  for (int r = 0; r < ph; r++)
    for (int c = pw - D; c < pw; c++)
      if (getbit(P, pw, ph, r, c))
        spill++;
  CK(0 == spill, "pixel spill %d", spill);
  for (int r = 0; r < ph; r++) {
    memset(R, 0, rbs);  // clear row buffer
    for (int b, c = 0; c < pw; c++) {  // set output row bits
      if      (D <= c && getbit(P, pw, ph, r, c-2))  b = getbit(K, kw, kh, r, c-2);  // shift P region in K rightward
      else if (          getbit(P, pw, ph, r, c  ))  b = randbit();                  // fill in gaps due to shift
      else                                           b = getbit(K, kw, kh, r, c);    // copy K background
      setbit(R, pw, c, b);
    }
    CK(fwrite(R, sizeof *R, rbs, stdout) == rbs, ES, EA);
  }
  free(R);
}

int main(int argc, char *argv[]) {
  CK(4 == argc, "bad argc %d", argc);
  switch (argv[1][0]) {
    case 'k':  randpbm(atoi(argv[2]), atoi(argv[3]));  break;
    case 'e':  makerds(     argv[2],       argv[3] );  break;
    default:   CK(0, "bad op '%c'", argv[1][0]);  break;
  }
  return 0;
}

